#include <mem.h>
#include "2D_RLE.h"

// 2D Run-Lenght encoding routines 
// Copyright (c) 2013 by demmel products gmbh - www.demmel.com

// ------------- variables used for decoding ------------- 

typedef enum
{
	os_init,
	os_horz_rep,
	os_horz_rep_2,
	os_uncompressed,
	os_uncompressed_2
} out_state_t;

static word *FPrevLine;
static word *FOutPntr;
static word *FOutLine;
static word FColor;
static int FImageWidth;
static int FBlockRep;
static int FOutCount;
static out_state_t FOutState;

// ------------- macros ------------- 

// color bytes of an iLCD scanline are sent in the order low byte first, high byte next
#define FirstColorByte(b)		(b)
#define SecondColorByte(b)		(((word) (b)) << 8)

//---------------------------------------------------------------------------
void InitDecode(word *act_scanline, word *prev_line, int width)
{
	FPrevLine = prev_line;
	FImageWidth = width;
	FOutLine = FOutPntr = act_scanline;
	FOutCount = 0;
	FOutState = os_init;
	FBlockRep = 0;
}
//---------------------------------------------------------------------------
// Note: DecodeByte() takes care to not overwrite the buffer 'FOutLine'
// Returns # of word values (pixels) written to FOutPntr (-1 in case of error)
int DecodeByte(byte b)
{
	int pixel_written = FOutPntr - FOutLine;			// # of pixels already written
	int pixel_left = FImageWidth - pixel_written;   // # of pixels which can be written
	
	int cnt;			// intermediate count
	int rv = 0;		// return value (# of pixels written)
	
	switch (FOutState)
	{
		case os_init:
			if ((b & BLOCK_REPETITION_FLAG) == BLOCK_REPETITION_FLAG)
			{
				FBlockRep = (b & 0x3F) + 1;
			}
			else if (b & HORZ_REPETITION_FLAG)
			{
				FOutCount = ((b & 0x3F) + 1) + FBlockRep * 64; 
				FBlockRep = 0;
				FOutState = os_horz_rep;
			}
			else if (b & VERT_REPETITION_FLAG)
			{
				if (FPrevLine == NULL)
				{
					FOutState = os_init;

					// encountered an error
					return -1;
				}
			
				FOutCount = ((b & 0x3F) + 1) + FBlockRep * 64;
				FBlockRep = 0;

				// we assume FPrevLine contains sensefull data  

				cnt = min(pixel_left, FOutCount);

				memcpy(FOutPntr, FPrevLine + pixel_written, cnt * 2); 
				FOutPntr += cnt;

				rv = cnt;
					
				FOutState = os_init;
			}       
			else
			{
				FOutCount = ((b & 0x3F) + 1) + FBlockRep * 64;
				FBlockRep = 0;
				FOutState = os_uncompressed;
			}
			break;

		case os_horz_rep:
			FColor = FirstColorByte(b);
			FOutState = os_horz_rep_2;
			break;

		case os_horz_rep_2:				
			FColor |= SecondColorByte(b);

			rv = FOutCount;

			cnt = min(pixel_left, FOutCount);
			
			while (cnt--)
				*FOutPntr++ = FColor;
				
			FOutState = os_init;
			break;

		case os_uncompressed:
			FColor = FirstColorByte(b);
			FOutState = os_uncompressed_2;
			break;
				
		case os_uncompressed_2:
			FColor |= SecondColorByte(b);

			if (pixel_left > 0)
				*FOutPntr++ = FColor;

			rv = 1;

			if (--FOutCount <= 0)
				FOutState = os_init;
			else					
				FOutState = os_uncompressed;
			break;
	}                        

	return rv;
}
//---------------------------------------------------------------------------
/*
	 'data' contains compressed data (with 'count' bytes), 'act_scanline' and 
	 'prev_scanline' is the actual (to be written to) and previous (already written 
	 on last call) scan line made up with 'width' pixel (word values, most significant
	byte is upper byte, e.g. the color red has a value of 0xF800).
	
	Decoding is done to the word array 'act_scanline' which needs to can hold a 
	minimum size of 'width' words.
	
	Function returns false in case of error.
*/	
//---------------------------------------------------------------------------
bool RunLengthDecode(byte *data, int count, word *act_scanline, word *prev_scanline, int width)
{
	int cnt;
	int pix_out_count = 0;
	
	if (width > MAX_SCANLINE_WIDTH || width < 0 || count < 1)
		return false;
		
	InitDecode(act_scanline, prev_scanline, width);

	while (count-- > 0)
	{
		cnt = DecodeByte(*data++);

		if (cnt < 0)
		{
			// error in data
			return false;
		}                  
		
		pix_out_count += cnt;

		if (pix_out_count > width)
		{
			// this should *never* happen !!
			return false;
		}
	}

	return pix_out_count == width; 
}
//---------------------------------------------------------------------------
// returns # of same pixels beginning from start (modifies start accordingly)
// returns -1 if no match found at all
static int MatchPrevLine(word *act_line, word *prev_line, int *start, int width)
{
	int x;
	int match_start, match_begin, match_count;
	
	if (prev_line == NULL)
		return -1;

	match_begin = -1;
	match_count = 0;
	match_start = *start;
		
	if (match_start < 0)
		match_start = 0;

	for (x = match_start; x < width; x++)
	{
		if (act_line[x] == prev_line[x])
		{
			if (match_begin < 0)
				match_begin = x;				

			match_count++;				
		}
		else
		{
			if (match_begin >= 0)
				break;
		}
	}

	if (match_count > 0)
		*start = match_begin;
	else
		return -1;		
		
	return match_count;		
}
//---------------------------------------------------------------------------
// returns # of bytes written to out_data.
// out_data mus be able to hold at least 'width' * 2 + 2 byte.
//	Note: if 'prev_scanline' is NULL, only 1D run-length encoding is carried out.
int RunLengthEncode(byte *out_data, word *act_line, word *prev_line, int width)
{
	// The iLCD controller expects to get low byte of color first, high byte next
	#define ColorOut(c) { *out_pntr++ = c & 0xFF; *out_pntr++ = c >> 8; }

	byte *ctrl_seq_pntr = out_data;
	byte *block_rep_pntr = NULL;
	byte *out_pntr = ctrl_seq_pntr + 1;
	int in_count, same_color_count, diff_color_count, match_start, match_count;
	word last_color;
	word actual_color;
	int x = 0;

	same_color_count = 0;
	diff_color_count = 0;
	in_count = width;

	match_start = 0;
	match_count = MatchPrevLine(act_line, prev_line, &match_start, width);

	if (match_count > 0 && match_start == 0)
   {
		if (match_count > 64)
			*ctrl_seq_pntr++ = BLOCK_REPETITION_FLAG | (((match_count - 1) / 64) - 1);
					
		*ctrl_seq_pntr++ = VERT_REPETITION_FLAG | ((match_count - 1) % 64);

		x += match_count; 

		if (x >= width)
			return ctrl_seq_pntr - out_data;
		
		in_count -= match_count;

		out_pntr = ctrl_seq_pntr + 1;
		
		match_count = 0;
	}

	actual_color = last_color = act_line[x++];
	ColorOut(last_color);
	*ctrl_seq_pntr = 0;

	if (match_count == 0)
	{
		match_start = x;
		match_count = MatchPrevLine(act_line, prev_line, &match_start, width);
	}			

	while (--in_count > 0)
	{
		if (x >= width)		// this should *never* happen !!
			break;

		if (match_count == 0)
		{
			match_start = x;
			match_count = MatchPrevLine(act_line, prev_line, &match_start, width);
		}			

		if (match_count > 0 && match_start == x)
		{
			if (diff_color_count)
			{
				*ctrl_seq_pntr = diff_color_count;
				ctrl_seq_pntr = out_pntr;
				
				block_rep_pntr = NULL;
				diff_color_count = 0;
			}      
			else if (same_color_count)
			{
				if (same_color_count > 63)
				{
					*ctrl_seq_pntr++ = BLOCK_REPETITION_FLAG | ((same_color_count / 64) - 1);
				}
				
				*ctrl_seq_pntr = HORZ_REPETITION_FLAG | (same_color_count % 64);
				out_pntr = ctrl_seq_pntr + 1;
				ColorOut(last_color);
				ctrl_seq_pntr = out_pntr++;
				*ctrl_seq_pntr = 0;

				// output new color
				ColorOut(actual_color);

				same_color_count = 0;
				diff_color_count = 0;
			}
			else if (block_rep_pntr != NULL && !same_color_count)
			{
				// We had a block of exactly 64 bytes (or a multiple of)
				// diff_color_count before. We must act like not having found 
				// the same color as there is no repetition count value of 0

				last_color = actual_color = act_line[x++];
				ColorOut(actual_color);
				diff_color_count++;

				match_count--;
				match_start++;

				if (match_count > 0)
				{
					*ctrl_seq_pntr = diff_color_count;
					ctrl_seq_pntr = out_pntr;
				
					block_rep_pntr = NULL;
					diff_color_count = 0;
				}
			}
			else
			{
				// special case: we had only one different pixel before the match
					
				out_pntr += 3;
				ctrl_seq_pntr += 3;
				*ctrl_seq_pntr = 0;
			}

			if (match_count > 0 && match_start == x)
			{
				if (match_count > 64)
					*ctrl_seq_pntr++ = BLOCK_REPETITION_FLAG | (((match_count - 1) / 64) - 1);
					
				*ctrl_seq_pntr++ = VERT_REPETITION_FLAG | ((match_count - 1) % 64);

				x += match_count; 

				if (x >= width)
					return ctrl_seq_pntr - out_data;
				
				in_count -= match_count;

				out_pntr = ctrl_seq_pntr + 1;
				match_count = 0;

				actual_color = last_color = act_line[x++];
				ColorOut(last_color);
				*ctrl_seq_pntr = 0;
				
				continue;
			}
		}
			
		if ((actual_color = act_line[x++]) == last_color)
		{
			if (diff_color_count)
			{
				*ctrl_seq_pntr = diff_color_count - 1;

				out_pntr--;
				ctrl_seq_pntr = out_pntr - 1;
				same_color_count = 1;
				
				block_rep_pntr = NULL;
				diff_color_count = 0;
				*ctrl_seq_pntr = 0;
			}      
			else
			{
				if (block_rep_pntr != NULL && !same_color_count)
				{
					// We had a block of exactly 64 bytes (or a multiple of) 
					// diff_color_count before. We must act like not having found 
					// the same color as there is no repetition count value of 0

					ColorOut(actual_color);
					diff_color_count++;
				}
				else
					same_color_count++;
			}
		}
		else
		{
			if (same_color_count)
			{
				if (same_color_count > 63)
				{
					*ctrl_seq_pntr++ = BLOCK_REPETITION_FLAG | ((same_color_count / 64) - 1);
				}
				
				*ctrl_seq_pntr = HORZ_REPETITION_FLAG | (same_color_count % 64);
				out_pntr = ctrl_seq_pntr + 1;
				ColorOut(last_color);
				ctrl_seq_pntr = out_pntr++;
				*ctrl_seq_pntr = 0;

				// output new color
				ColorOut(actual_color);

				same_color_count = 0;
				diff_color_count = 0;
			}
			else
			{
				if (++diff_color_count > 63)
				{
					if (block_rep_pntr != NULL)
					{
						(*block_rep_pntr)++;
					}
					else
					{
						memmove(ctrl_seq_pntr + 1, ctrl_seq_pntr, 64 * 2 + 1);
						block_rep_pntr = ctrl_seq_pntr++;
						*block_rep_pntr = BLOCK_REPETITION_FLAG;
						out_pntr++;
					}

					diff_color_count = 0;
				}

				ColorOut(actual_color);
			}           
		}

		last_color = actual_color;
	}

	if (diff_color_count)
	{
		*ctrl_seq_pntr = diff_color_count;
	}
	else if (same_color_count)
	{
		if (same_color_count > 63)
		{
			*ctrl_seq_pntr++ = BLOCK_REPETITION_FLAG | ((same_color_count / 64) - 1);
		}
				
		*ctrl_seq_pntr = HORZ_REPETITION_FLAG | (same_color_count % 64);
		out_pntr = ctrl_seq_pntr + 1;
		ColorOut(last_color);
	}
	
	return out_pntr - out_data;

	#undef ColorOut
}
//---------------------------------------------------------------------------
