//---------------------------------------------------------------------------
#include "windows.h"

#include "serial.h"
#include "errmsgs.h"

#define DEFAULT_RX_TIMEOUT	500 	// ms
#define DEFAULT_TX_TIMEOUT	100	// ms

//---------------------------------------------------------------------------

static bool fIsOpen = false;
static dword fBaudRate = 115200;
static HANDLE fHandle = INVALID_HANDLE_VALUE;
static DCB fDCB;
static dword curr_rx_tout = 0;
static dword curr_tx_tout = 0;

//---------------------------------------------------------------------------
// Set the timeout values of an already opened port
bool Serial_SetTimeouts(dword rx_tout, dword tx_tout)
{
	COMMTIMEOUTS fTOuts;

	if (!fIsOpen || fHandle == INVALID_HANDLE_VALUE)
	{
		// "Set Timeout error: Port is not open"
		return Error_SetErr(SER_ERR_TOUT_NOT_OPEN);
	}

	// avoid overhead
	if (rx_tout == curr_rx_tout && tx_tout == curr_tx_tout)
		return true;

	// Wait 'tx_tout' ms seconds for sending until system complains
	fTOuts.WriteTotalTimeoutMultiplier = 1;
	fTOuts.WriteTotalTimeoutConstant = tx_tout;

	// set read interval timeout to 100 ms
	fTOuts.ReadIntervalTimeout = 100;
	fTOuts.ReadTotalTimeoutMultiplier = 1;
	// Wait 'rx_tout' ms until we give up
	fTOuts.ReadTotalTimeoutConstant = rx_tout;

	if (!SetCommTimeouts(fHandle, &fTOuts))
	{
		// "Set CommTimeout error: " + System message
		return Error_SetErr(SER_ERR_TOUT);
	}

	curr_rx_tout = rx_tout;
	curr_tx_tout = tx_tout;
	
	return true;
}
//---------------------------------------------------------------------------
// Set the serial port values of an already opened port
bool Serial_Config(dword baud_rate, bool no_cts)
{
	if (!fIsOpen || fHandle == INVALID_HANDLE_VALUE)
	{
		// "Config error: Port is not open"
		return Error_SetErr(SER_ERR_CONFIG_NOT_OPEN);
	}

	// Fill dcb with current values
	if (!GetCommState(fHandle, &fDCB))
	{
		// "Config error: " + System message
		return Error_SetErr(SER_ERR_CONFIG);
	}

	fDCB.BaudRate = baud_rate;
	fDCB.fBinary = true;
	fDCB.fParity = false;
	fDCB.fOutxCtsFlow = no_cts ? false: true;
	fDCB.fOutxDsrFlow = false;
	fDCB.fDtrControl = DTR_CONTROL_ENABLE;
	fDCB.fDsrSensitivity = false;
	fDCB.fOutX = false;
	fDCB.fInX = false;
	fDCB.fErrorChar = false;
	fDCB.fNull = false;
	fDCB.fRtsControl = RTS_CONTROL_ENABLE;
	fDCB.fAbortOnError = false;
	fDCB.ByteSize = 8;
	fDCB.Parity = NOPARITY;
	fDCB.StopBits = TWOSTOPBITS;

	fBaudRate = baud_rate;

	if (!SetCommState(fHandle, &fDCB))
	{
		// "Set CommState error: " + System message
		return Error_SetErr(SER_ERR_COMSTATE);
	}

	SetupComm(fHandle, 4096, 4096);

	// Set default timeout
	return Serial_SetTimeouts(DEFAULT_RX_TIMEOUT, DEFAULT_TX_TIMEOUT);
}
//---------------------------------------------------------------------------
// Open serial port - note that the applciation has to wait until data is sent
// and RX data is polled. A better approach would be to use overlapped i/o and
// seperate threads for transmitting/receiving
bool Serial_Open(char *port, dword baud_rate, bool no_cts)
{
	char ser_port[100 + 1];
	char com_name[100 + 1];
	
	if (fIsOpen)
	{
		// "Port is already open"
		return Error_SetErr(SER_ERR_ALREADY_OPEN);
	}

	strncpy(ser_port, port, 100);
	ser_port[100] = '\0';
	strlwr(ser_port);

	if (strncmp(ser_port, "com", 3) != 0)
	{
		// "Invalid serial port name specified"
		return Error_SetErr(SER_INVALID_NAME);
	}

	if (ser_port[strlen(ser_port) - 1] == ':')
		ser_port[strlen(ser_port) - 1] = '\0';      

	strcpy(com_name, "\\\\.\\");		
	strncat(com_name, ser_port, 100 - strlen(com_name));
	com_name[100] = '\0';
								
	fHandle = CreateFile(com_name,
		GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, NULL);

	if (fHandle == INVALID_HANDLE_VALUE)
	{
		// "Open error: " + System message
		return Error_SetErr(SER_ERR_CANT_OPEN);
	}

	if (GetFileType(fHandle) != FILE_TYPE_CHAR)
	{
		CloseHandle(fHandle);
		fHandle = INVALID_HANDLE_VALUE;

		// "Open error: Device is not a character device"
		return Error_SetErr(SER_ERR_OPEN_NO_CHAR_DEV);
	}

	fIsOpen = true;

	return Serial_Config(baud_rate, no_cts);
}
//---------------------------------------------------------------------------
// Close serial port if it is open
bool Serial_Close(void)
{
	bool rv;

	curr_rx_tout = 0;
	curr_tx_tout = 0;

	if (!fIsOpen || fHandle == INVALID_HANDLE_VALUE)
	{
		// "Close error: Device was not open"
		return Error_SetErr(SER_ERR_CLOSE_NOT_OPEN);
	}

	rv = CloseHandle(fHandle);
	fHandle = INVALID_HANDLE_VALUE;
	fIsOpen = false;

	if (!rv)
	{
		// "Close error: Unable to close serial port"
		return Error_SetErr(SER_ERR_CANT_CLOSE);
	}

	return true;
}
//---------------------------------------------------------------------------
// Transmit buffer 'data' containing 'size' bytes
// The buffer can include 0-characters as well
bool Serial_Transmit(unsigned char *data, dword size)
{
	dword size_sent;

	if (!fIsOpen)
	{
		// "Transmit error: Port is not open"
		return Error_SetErr(SER_ERR_TX_NOT_OPEN);
	}

	if (!WriteFile(fHandle, data, size, &size_sent, NULL))
	{
		// "Transmit error: " + System message
		return Error_SetErr(SER_ERR_TX);
	}

	if (size != size_sent)
	{
		// "Transmit error: Could not send data"
		return Error_SetErr(SER_ERR_TX_SIZE);
	}

	return true;
}
//---------------------------------------------------------------------------
// Get a byte from the LCD panel - return -1 if no data is available within
// the time setup in Config()
bool Serial_ReceiveChar(unsigned char *c)
{
	dword bytes_got = 0;

	if (!fIsOpen)
	{
		// "Receive error: Port is not open"
		return Error_SetErr(SER_ERR_RX_NOT_OPEN);
	}

	if (ReadFile(fHandle, c, 1, &bytes_got, NULL))
	{
		if (bytes_got)
			return true;

		// "No data to read"
		return Error_SetErr(SER_ERR_RX_NO_DATA);
	}

	// "Read error: " + System message
	return Error_SetErr(SER_ERR_RX);
}
//---------------------------------------------------------------------------
bool Serial_IsOpen()
{
	return fIsOpen;
}
//---------------------------------------------------------------------------
dword Serial_GetBaudRate()
{
	return fBaudRate;
}
//---------------------------------------------------------------------------


